"""MISC FUNCTIONS
You shouldn't need to tweak these much if at all
"""

import sys
import os
import subprocess
import yaml
import click
from shutil import copyfile
from time import localtime, strftime


class OrderedCommands(click.Group):
    """This class will preserve the order of subcommands, which is useful when printing --help"""
    def list_commands(self, ctx: click.Context):
        return list(self.commands)


def snake_base(rel_path):
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), rel_path)


def print_version():
    with open(snake_base('snaketool.VERSION'), 'r') as f:
        version = f.readline()
    echo_click('\n' + 'snaketool version ' + version + '\n')


def echo_click(msg, log=None):
    click.echo(msg, nl=False, err=True)
    if log:
        with open(log, 'a') as l:
            l.write(msg)


def print_citation():
    with open(snake_base('snaketool.CITATION'), 'r') as f:
        for line in f:
            echo_click(line)


def msg(err_message, log=None):
    tstamp = strftime('[%Y:%m:%d %H:%M:%S] ', localtime())
    echo_click(tstamp + err_message + '\n', log=log)


def msg_box(splash, errmsg=None, log=None):
    msg('-' * (len(splash) + 4), log=log)
    msg(f'| {splash} |', log=log)
    msg(('-' * (len(splash) + 4)), log=log)
    if errmsg:
        echo_click('\n' + errmsg + '\n', log=log)


def default_to_ouput(ctx, param, value):
    """Callback for --configfile; place value in output directory unless specified"""
    if param.default == value:
        return os.path.join(ctx.params['output'], value)
    return value


def copy_config(local_config, system_config=snake_base(os.path.join('config', 'config.yaml')), log=None):
    if not os.path.isfile(local_config):
        if len(os.path.dirname(local_config)) > 0:
            os.makedirs(os.path.dirname(local_config), exist_ok=True)
        msg(f'Copying system default config to {local_config}', log=log)
        copyfile(system_config, local_config)
    else:
        msg(f'Config file {local_config} already exists. Using existing config file.', log=log)


def read_config(file):
    with open(file, 'r') as stream:
        _config = yaml.safe_load(stream)
    return _config


def write_config(_config, file, log=None):
    msg(f'Writing runtime config file to {file}', log=log)
    with open(file, 'w') as stream:
        yaml.dump(_config, stream)


"""RUN A SNAKEFILE
Hopefully you shouldn't need to tweak this function at all.
- You must provide a Snakefile, all else is optional
- Highly recommend supplying a configfile and the default snakemake args"""


def run_snakemake(configfile=None, snakefile_path=None, merge_config=None, threads=1, use_conda=False,
                  conda_prefix=None, snake_default=None, snake_args=[], log=None):
    """Run a Snakefile"""
    snake_command = ['snakemake', '-s', snakefile_path]

    # if using a configfile
    if configfile:
        # copy sys default config if not present
        copy_config(configfile, log=log)

        # read the config
        snake_config = read_config(configfile)

        # merge in command line config if provided
        if merge_config:
            snake_config.update(merge_config)

            # update config file for Snakemake execution
            write_config(snake_config, configfile, log=log)

        snake_command += ['--configfile', configfile]

        # display the runtime configuration
        msg_box('Runtime config', errmsg=yaml.dump(snake_config, Dumper=yaml.Dumper), log=log)

    # add threads
    if not '--profile' in snake_args:
        snake_command += ['--jobs', threads]

    # add conda args if using conda
    if use_conda:
        snake_command += ['--use-conda']
        if conda_prefix:
            snake_command += ['--conda-prefix', conda_prefix]

    # add snakemake default args
    if snake_default:
        snake_command += snake_default

    # add any additional snakemake commands
    if snake_args:
        snake_command += list(snake_args)

    # Run Snakemake!!!
    snake_command = ' '.join(str(s) for s in snake_command)
    msg_box('Snakemake command', errmsg=snake_command, log=log)
    if not subprocess.run(snake_command, shell=True).returncode == 0:
        msg('ERROR: Snakemake failed', log=log)
        sys.exit(1)
    else:
        msg('Snakemake finished successfully', log=log)
    return 0


import os
from setuptools import setup, find_packages

def get_version():
    with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'snaketool', 'snaketool.VERSION')) as f:
        return f.readline().strip()

CLASSIFIERS = [
    "Environment :: Console",
    "Environment :: MacOS X",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: MIT license",
    "Natural Language :: English",
    "Operating System :: POSIX :: Linux",
    "Operating System :: MacOS :: MacOS X",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Topic :: Scientific/Engineering :: Bio-Informatics",
]

setup(
 name='snaketool',
 packages=find_packages(),
 url='',
 python_requires='>=3.7',
 description="Snakemake-powered commandline tool to do a thing.",
 version=get_version(),
 author="Michael Roach",
 author_email="example@email.com",
 py_modules=['snaketool'],
 install_requires=["snakemake==7.14.0",
                   "pyyaml==6.0",
                   "Click==8.1.3"],
 entry_points={
  'console_scripts': [
    'snaketool=snaketool.__main__:main'
  ]},
 include_package_data=True,
)



"""
Entrypoint for snaketool

Check out the wiki for a detailed look at customising this file:
https://github.com/beardymcjohnface/Snaketool/wiki/Customising-your-Snaketool
"""

import os
import click

from .util import snake_base, print_version, default_to_ouput, copy_config, run_snakemake, OrderedCommands, print_citation


def common_options(func):
    """Common command line args
    Define common command line args here, and include them with the @common_options decorator below.
    """
    options = [
        click.option('--output', help='Output directory', type=click.Path(dir_okay=True, writable=True, readable=True),
                     default='snaketool.out', show_default=True),
        click.option('--configfile', default='config.yaml', show_default=False, callback=default_to_ouput,
                     help='Custom config file [default: (outputDir)/config.yaml]'),
        click.option('--threads', help='Number of threads to use', default=1, show_default=True),
        click.option('--use-conda/--no-use-conda', default=True, help='Use conda for Snakemake rules',
                     show_default=True),
        click.option('--conda-prefix', default=snake_base(os.path.join('workflow', 'conda')),
                     help='Custom conda env directory', type=click.Path(), show_default=False),
        click.option('--snake-default', multiple=True,
                     default=['--rerun-incomplete', '--printshellcmds', '--nolock', '--show-failed-logs'],
                     help="Customise Snakemake runtime args", show_default=True),
        click.option('--log', default='snaketool.log', callback=default_to_ouput, hidden=True),
        click.argument('snake_args', nargs=-1)
    ]
    for option in reversed(options):
        func = option(func)
    return func


@click.group(cls=OrderedCommands,context_settings=dict(help_option_names=["-h", "--help"]))
def cli():
    """For more options, run:
    snaketool command --help"""
    pass


help_msg_extra = """
\b
CLUSTER EXECUTION:
snaketool run ... --profile [profile]
For information on Snakemake profiles see:
https://snakemake.readthedocs.io/en/stable/executing/cli.html#profiles
\b
RUN EXAMPLES:
Required:           snaketool run --input [file]
Specify threads:    snaketool run ... --threads [threads]
Disable conda:      snaketool run ... --no-use-conda 
Change defaults:    snaketool run ... --snake-default="-k --nolock"
Add Snakemake args: snaketool run ... --dry-run --keep-going --touch
Specify targets:    snaketool run ... all print_targets
Available targets:
    all             Run everything (default)
    print_targets   List available targets
"""


@click.command(epilog=help_msg_extra, context_settings=dict(help_option_names=["-h", "--help"], ignore_unknown_options=True))
@click.option('--input', '_input', help='Input file/directory', type=str, required=True)
@common_options
def run(_input, output, log, **kwargs):
    """Run snaketool"""
    # Config to add or update in configfile
    merge_config = {
        'input': _input,
        'output': output,
        'log': log}

    # run!
    run_snakemake(
        snakefile_path=snake_base(os.path.join('workflow', 'Snakefile')),   # Full path to Snakefile
        merge_config=merge_config,
        log=log,
        **kwargs
    )


@click.command()
@common_options
def config(configfile, **kwargs):
    """Copy the system default config file"""
    copy_config(configfile)


@click.command()
def citation(**kwargs):
    """Print the citation(s) for this tool"""
    print_citation()


cli.add_command(run)
cli.add_command(config)
cli.add_command(citation)


def main():
    print_version()
    cli()


if __name__ == '__main__':
    main()
